/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Page management
*  SCOPE :              Page management
*  FILE :               page.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#ifndef __page_HXX
#include "page.hxx"
#endif

#ifndef __var_H
#include "var.h"
#endif

#ifndef __define_H
#include "define.h"
#endif

#ifndef __cluster_HXX
#include "cluster.hxx"
#endif

#ifndef __smartp_HXX
#include "smartp.hxx"
#endif

#ifndef __virtualdbserver_HXX
#include "virtualdbserver.hxx"
#endif

#include "alphadecl.h"


/*****************************************************************************/


/*= EXTERNAL FUNCTION =======================================================*/
void Page::init(Cluster* clu, PID pid, int nbPage) 
/*=
[DESCRIPTION :]
  Initialise explicitly the page. Because a page couldn't be build by 
  traditionnal C++ constructor, we use an explicit call to init method. It 
  prepare the page to be a short page. However, the init method is used to
  build long pages. The drawback is that we lose space for object directory
  in a long page header. 

  (IN)
    <Cluster* clu>
      => Cluster where the page is created. Use to define lock mode.
    <PID pid>
      => Space must be allocated in persistant store before init. So we know 
      => the pid of the page
    <int [1..PAGE_IN_VOL] nbPage>
      => Define the size of the block. Used only for long object

[DIAGNOSTIC :]
  <YOO_ERR_DBG> 
    => If nbPage is not in the correct range

[CALL :]
  LocalTrans
    => Call when user allocates pages in a transaction
=*/
{
#ifdef DEBUG
  if ((nbPage > PAGE_IN_VOL) || (nbPage < 1)) {
    yoo_error(YOO_ERR_DBG,"(Page::init) : init out of range");
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif

  this->_usedSpace = 0 ;
  this->_pgAfter = nbPage - 1 ; 
  this->_cluster = clu->getOid() ;
  this->_pid = pid ;
  for (int i = 0; i < MAX_DIR_PG; i++) {
    this->_tabObj[i] = FREE_SLOT ;
  }
  this->_firstFreeSlot = 0 ;
  this->_tab[0].slot()->setFree(SLOT_IN_PG) ;
  this->_tab[0].slot()->next() = -1 ;

  // Case of a long object
  if (nbPage > 1) {
    this->_tabObj[0] = 0 ; 
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
Cluster* Page::getCluster()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Ref<Cluster> clu(this->_cluster) ; 
  return clu ;
}


/*= EXTERNAL FUNCTION =======================================================*/
ObjectX* Page::allocObject(int ss) 
/*=
[DESCRIPTION :]
  Allocation of an object with size ss in slot. There are four reasons for 
  which the page could not allocate the space. 
  First, if the page is a long page, allocation of short object is prohibited. 
  This case could append when user want to allocate an object close to another.
  Automatic attempt is done in the page of the desired neighbourg. 
  Second, if the page has a total free space which is less than needed space.
  Third, if the object directory is full. Your DB must be garbaged
  Fourth, if there is not enough contigous slots. This case is reduced because 
  page are compacted when receive on client.

  (IN)
    <int [1..SLOT_IN_PG] ss>
      => Size of the needed space in slot
  (OUT)
    ObjectX*
      => Virtual memory pointer on the allocated space. 

[DIAGNOSTIC :]
  <YOO_ERR_DBG> 
    => If ss is not in the correct range
    => If consistency test is false

  (RETURN)
    <NULL>
      => If the object could not be allocated

[CALL :]
  BlocAlloc
    => Call when we find the page in the blocAllocator of a cluster
  Cluster
    => Call before to insert in blocAllocator when we allocate a new page 
  ObjectX
    => Call in the new operator and when the object change its size
=*/
{
#ifdef DEBUG 
  if ((ss > SLOT_IN_PG) || (ss <= 0)) {
    yoo_error(YOO_ERR_DBG, "=> Strange size for an object to alloc") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif

  // If the size could not be allocated in the page, return NULL
  // If the page is a long page, return NULL
  if ((ss > this->freeSpace()) || (this->size() > 1)) {
    return NULL ;
  }

  OID oid = this->_pid ;
  // Allocate an entry in the page directory table
  oid.objNo = this->allocTab() ;
  if (oid.objNo == MAX_DIR_PG) {
    return NULL ;
  } 

  int curr = this->_firstFreeSlot ;
  short* prev = &this->_firstFreeSlot ;

  // Scan every free blocs
  while (curr != -1) {
    int diff = this->_tab[curr].slot()->getSlotSize() - ss ;
    // if current free bloc fits
    if (diff >= 0) {
      this->_usedSpace += ss ;
      // Test if the free bloc is used completly or not
      if (diff == 0) {
        *prev = this->_tab[curr].slot()->next() ;
      } else {
	// insert unused block in the free list block
        *prev = curr + ss ;
        this->_tab[*prev].slot()->setFree(diff) ;
        this->_tab[*prev].slot()->next() = this->_tab[curr].slot()->next() ;
      }
      // initialise object
      this->_tabObj[oid.objNo] = curr ;
      this->object(oid.objNo)->initObj(oid, ss) ;
      return this->object(oid.objNo) ;
    }
    prev = &this->_tab[curr].slot()->next() ;
    curr = this->_tab[curr].slot()->next() ;
  }

#ifdef DEBUG
  if (this->consistency() == FALSE) {
    yoo_error(YOO_ERR_DBG, "Page inconsistent") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif
  return NULL ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void Page::freeObject(OID oid) 
/*=
[DESCRIPTION :]
  Delete the object from the page. If the object is a long object, we free all
  the pages.In the case of a short object, we free the corresponding slots and 
  try to compact with previous and next slots if there are free. By defining 
  NO_DANGLING,it is possible to avoid the reallocation of the OID. The drawback
  is that pages directory will grow up and possibly be full. So you must
  sometimes garbage your database explicitly with DBA tools.
  If the page has no more used space, we free the page. If not, we try to
  compact with previous or next free block, unlink them and relink at the end
  of the compaction list.

  (IN)
    <OID oid>
      => OID of the object we want to free

[DIAGNOSTIC :]
  <YOO_ERR_DBG> 
    => If OID is not consistent
    => If consistency test is false

[CALL :]
  ObjectX
    => Call in the delete operator and when the object change its size
=*/
{
  if ( (this->_tabObj[oid.objNo] == FREE_SLOT) ||
       (this->_tabObj[oid.objNo] == DEL_SLOT) ) {
    yoo_error(YOO_ERR_DBG, "Attempt to free an invalid object") ;
    yoo_exit(YOO_ERR_DBG) ;
  }

  // Long object page 
  if (this->_pgAfter > 0) {
    int sz = this->size() ;
    PID currPid = this->_pid ;
    for (int i = 0; i < sz; i++) {
      this[i]._usedSpace = 0 ;
      this[i]._pgAfter = 0 ;
      YOO_VirtualMemory->findPageHdl(currPid)->deleted() ;
      currPid += 1 ;
    }
    return ;
  }

  int slot = this->_tabObj[oid.objNo] ;
#ifdef NO_DANGLING  
  this->_tabObj[oid.objNo] = DEL_SLOT ; // No reallocation of the OID
#else  
  this->_tabObj[oid.objNo] = FREE_SLOT ; // Possible reallocation of the OID
#endif
  int ss = this->_tab[slot].slot()->getSlotSize() ;
  this->_usedSpace -= ss ;

  PageHdlMem* pd = YOO_VirtualMemory->findPageHdl(this->_pid) ;
  pd->remove() ;
  
  if (this->_usedSpace == 0) {
    pd->deleted() ;
    return ;
  }  

  // Scan free blocs for local recompaction
  int curr = this->_firstFreeSlot ;
  short* prev = &this->_firstFreeSlot ;
  int currS ;
  while (curr != -1) {
    currS = this->_tab[curr].slot()->getSlotSize() ;
    // 
    if (curr < slot) {
      if ((curr + currS) == slot) {
        slot = curr ;
        ss += currS ;
        // unlink the free block
        *prev = this->_tab[curr].slot()->next() ;
        curr = this->_tab[curr].slot()->next() ;
        continue ;
      }
    } else if ((slot + ss) == curr) {
      ss += currS ;
      // unlink the free block
      *prev = this->_tab[curr].slot()->next() ;
      curr = this->_tab[curr].slot()->next() ;
      continue ;
    }
  
    prev = &this->_tab[curr].slot()->next() ;
    curr = this->_tab[curr].slot()->next() ;
  }

  // Link the compacted free block at the head of the free block list
  this->_tab[slot].slot()->setFree(ss) ;
  this->_tab[slot].slot()->next() = this->_firstFreeSlot ;
  this->_firstFreeSlot = slot ;

  pd->insertCluster() ;
  
#ifdef DEBUG
  if (this->consistency() == FALSE) {
    yoo_error(YOO_ERR_DBG, "Page inconsistent") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif  
}


/*= INTERNAL FUNCTION =======================================================*/
void Page::receive()
/*=
[DESCRIPTION :]
  Receive is used when we access the first time to the page. It is used only 
  on pages containing short object or on first page of a long object. On a 
  short page, we compact the page and call refresh method on every object.
  Refresh methods are methods generated at the interface level for each classes
  of the schema. It reinitialized virtual table pointer and virtual pointers on
  virtual base classes. For a long object, we only call refresh method. In any
  case, refresh method is called only if the object has a user defined type. 
  Refresh methods are static function member of each classes of a schema and 
  are stored in an array indexed by class identifier. The array is initialised 
  at the beginning of a Yooda session.

[DIAGNOSTIC :]
  <YOO_ERR_DBG> 
    => If page is empty because it is abnormal to receive an empty page
    => If consistency test is false

[CALL :]
  VirtualDBServer
    => Call when receive the page from a volume into the cache
=*/
{
#ifdef DEBUG
  if ( (this->_usedSpace == 0) && (this->_pgAfter == 0) ) {
    yoo_error(YOO_ERR_DBG, "Receive empty page") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif

  YOO_RECV_PHASE = TRUE ;
  
  // long object
  if (_pgAfter >= 1) {
    // long object occupied always entry '0' of the page directory
    if (this->_tab[0].slot()->isTyped()) {
      YOO_TabRefresh[this->_tab[0].slot()->getClass()](&this->_tab[0]) ;
    }
    YOO_RECV_PHASE = FALSE ;
    return ;
  }

  // Begin compaction at the first slot.
  int up = this->firstUsed(0) ;
  int fp = this->firstFree() ;
  fp = fp > up ? up : fp ;
  int up2 ;

  // heuristic to find a compacted page
  if ( (up == 0) && (fp = this->_usedSpace) ) {
    while (up < SLOT_IN_PG) {
      up2 = up + this->_tab[up].slot()->getSlotSize() ;
        // Call refresh method if the object has a user defined type
      if (this->_tab[up].slot()->isTyped()) {
	YOO_TabRefresh[this->_tab[up].slot()->getClass()](&this->_tab[up]) ;
      }
      up = this->firstUsed(up2) ;
    }
    YOO_RECV_PHASE = FALSE ;
    SYSCALLI(mprotect((caddr_t)this, PAGE_SIZE, PROT_READ)) ;    
    return ;
  }

  // prepare a reverse tab to refresh page directory after compaction
  unsigned short pagemap[SLOT_IN_PG] ; 
  for (int i = 0; i < SLOT_IN_PG; i++) {
    pagemap[i] = FREE_SLOT ;
  }

  while (up < SLOT_IN_PG) {
    up2 = up + this->_tab[up].slot()->getSlotSize() ;
    // move and call refresh
    this->move(up, fp) ;
    pagemap[up] = fp ;
    up = this->firstUsed(up2) ;
    fp = fp + this->_tab[fp].slot()->getSlotSize() ;
  }
  
  // refresh page directory
  for (i = 0; i < MAX_DIR_PG; i++) {
    if ((this->_tabObj[i] != FREE_SLOT) && 
	(this->_tabObj[i] != DEL_SLOT) && 
	(pagemap[this->_tabObj[i]] != FREE_SLOT)) {
      this->_tabObj[i] = pagemap[this->_tabObj[i]] ;
    }
  }

  // insert uniq free block in the free list block
  if (fp < SLOT_IN_PG) {
    this->_firstFreeSlot = fp ;
    this->_tab[fp].slot()->setFree(SLOT_IN_PG-fp) ;
    this->_tab[fp].slot()->next() = -1 ;
  } else {
    this->_firstFreeSlot = -1 ;
  }

  YOO_RECV_PHASE = FALSE ;
  SYSCALLI(mprotect((caddr_t)this, PAGE_SIZE, PROT_READ)) ;
  
#ifdef DEBUG
  if (this->consistency() == FALSE) {
    yoo_error(YOO_ERR_DBG, "Page inconsistent") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif
} ;


/*= INTERNAL FUNCTION =======================================================*/
unsigned short Page::firstFree() 
/*=
[DESCRIPTION :]
  Return the first free slot in the page. It searches it by scanning block from
  the first one until it founds a free one.

  <OUT>
    <int [0..SLOT_IN_PG]>
      => Position in the slot tab of the first free block

[DIAGNOSTIC :]
[CALL :]
  Call in the receive method.
=*/
{
  int size = 0 ;
  for (int i = 0; i < SLOT_IN_PG; i += size) {
    if (this->_tab[i].slot()->isFree()) {
      return i ;
    } else {
      size = this->_tab[i].slot()->getSlotSize() ;
    }
  }
  return i ;
}  


/*= INTERNAL FUNCTION =======================================================*/
unsigned short Page::firstUsed(unsigned short fp) 
/*=
[DESCRIPTION :]
  Return the first used slot from fp. It searches it by scanning block from
  the fp block one until it founds a used block.

  <OUT>
    <int [0..SLOT_IN_PG]>
      => Position in the slot tab of the first used block

[DIAGNOSTIC :]
[CALL :]
  Call in the receive method.
=*/
{
  int size ;
  for (int i = fp; i < SLOT_IN_PG; i += size) {
    if (this->_tab[i].slot()->isFree()) {
      size = this->_tab[i].slot()->getSlotSize() ;
    } else {
      return i ;
    }
  }
  return i ;
}  


/*= INTERNAL FUNCTION =======================================================*/
void Page::move(unsigned short op, unsigned short np) 
/*=
[DESCRIPTION :]
  Move the object from op to fp slot by slot. Initialise the object by refresh
  method if it is typed.

[DIAGNOSTIC :]
[CALL :]
  Call in the receive method.
=*/
{
  // Copy the object when destination is different from source
  if (op != np) {
    int blk = this->_tab[op].slot()->getSlotSize() ;
    // Careful opposite of memcpy
    memmove((char*)&this->_tab[np], (char*)&this->_tab[op], SLOT_SIZE*blk) ;
  }

  // Call refresh method if the object has a user defined type
  if (this->_tab[np].slot()->isTyped()) {
    YOO_TabRefresh[this->_tab[np].slot()->getClass()](&this->_tab[np]) ;
  }
}


/*= INTERNAL FUNCTION =======================================================*/
unsigned short Page::allocTab() 
/*=
[DESCRIPTION :]
  find a free indice in the indirection table. Note that in 'NO_DANGLING' mode
  deleted entries are not reused. This method is not useful on long object 
  because there entry is always 0. A better solution will be to treat long 
  object as short one for directory so we can intercept use of dangling OID on
  long object. 

  <OUT>
    <int [0..MAX_DIR_PG]>
      => Position in the directory of a free entry

[DIAGNOSTIC :]
[CALL :]
  Call in the allocObject method.

[DIAGNOSTIC :]
=*/
{
  for(int i = 0; (i < MAX_DIR_PG) && (_tabObj[i] != FREE_SLOT); i++); 
  return i ; 
} 


/*= INTERNAL FUNCTION =======================================================*/
int Page::consistency(PID addr, int* no, int* ndo, int* sz, int* objPerClass)
/*=
[DESCRIPTION :]
  Test the consistency of the page
[DIAGNOSTIC :]
=*/
{
  if (YOO_STAT == FALSE) {
    char* deb = getenv("DEBUG") ;
    if ((deb == NULL) || (strcmp(getenv("DEBUG"), "TRUE"))) {
      return TRUE ;
    }
  }

  if ((addr != NULL_PID) && (addr != _pid)) {
    yoo_error(YOO_ERR_DBG,"Inconsistent oid for a page") ;
    return FALSE ;
  }

  if (_pgAfter > 0) {
    if ((this->_usedSpace != 0) ||
	(this->_cluster == NULL_OID)) {
      yoo_error(YOO_ERR_DBG,"Strange header for a long page") ;
      return FALSE ;
    } else {
      return TRUE ;
    }
  }

  int curr = this->_firstFreeSlot ;
  if ( (curr >= SLOT_IN_PG) || (curr < -1) ) {
    yoo_error(YOO_ERR_DBG,"Page::consistency => wrong first free bloc") ;
    return FALSE ;
  }

  // Scan free list 
  int fSpace = 0 ;
  while (curr != -1) {
    fSpace += this->_tab[curr].slot()->getSlotSize() ;
    if (this->_tab[curr].slot()->isFree() == FALSE) {
      yoo_error(YOO_ERR_DBG,
		"Page::consistency => non free object in free list") ;
      return FALSE ;
    }
    curr = this->_tab[curr].slot()->next() ;
  }

  // Scan used object
  int uSpace = 0 ;
  for (int i = 0; i < MAX_DIR_PG; i++) {
    if (_tabObj[i] > DEL_SLOT) {
      yoo_error(YOO_ERR_DBG,"Page::consistency => Strange directory") ;
      return FALSE ;
    }
    if ((_tabObj[i] == DEL_SLOT) && (ndo != NULL)) (*ndo)++ ;
    if (_tabObj[i] < SLOT_IN_PG) {
      if (no != NULL) (*no)++ ;
      uSpace += _tab[_tabObj[i]].slot()->getSlotSize() ;
      OID oid = this->object(i)->getOid() ;
      int classIdent = this->object(i)->getClass() ;
      if (objPerClass != NULL) objPerClass[classIdent]++ ; 
      int ss = this->object(i)->getSlotSize() ;
      if ((classIdent <= FREEBLK) ||
	  (classIdent > YOO_LastClassId) ||
	  (ss <= 0) ||
	  (ss > SLOT_IN_PG) ||
	  (oid.objNo != i) || 
	  (oid.pid() != this->_pid)) {
	yoo_error(YOO_ERR_DBG,"Page::consistency => Strange object") ;
	return FALSE ;
      }
    }
  }

  // test total space
  if (uSpace != this->_usedSpace) {
    yoo_error(YOO_ERR_DBG,"Page::free => error in space count") ;
    return FALSE ;
  }

  if ( (fSpace + uSpace) != SLOT_IN_PG) {
    yoo_error(YOO_ERR_DBG,"Page::free => error in space count") ;
    return FALSE ;
  }

  if (sz != NULL) *sz += uSpace * SLOT_SIZE ;

  return TRUE ;
}



